/*=======================================================
  Author:   
  Version:  $Id$
  Comment:  
            
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"


/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "OptInlineIndir"     

/* added just in case */
extern void IndirInit()
{

}

/*=======================================================
=======================================================*/


/* added to return memory from indirect jump profiling */
extern void IndirExit()
{
    if( !HasOption(DISABLED_OPT,"Indir") ) {
	OptDupIndir(stdout);
    }
    free(aIndir);
}


/*=======================================================
=======================================================*/


extern void AfunStealCalleeRegister(INDEX ifun, REG r) 
{
    INDEX jbbl;
    INDEX iins;
    INDEX saveIns;

    FOREACH_FUN_BBL(ifun, jbbl) {
	FOREACH_BBL_INS(jbbl, iins) {
	    if (AINS_TYPE(iins) != IT_LDA) continue;
	    if (AINS_REG_C(iins) != REG_SP) continue;
	    if (AINS_DATA(iins) < 0) {
		/* growing the stack */
		AINS_DATA(iins) -= 8;
		/* add in a store of callee save register, so 
		 *  we can use it for computing values in the loop 
		 */
		/* stq r, (sp)  */
		saveIns = AinsNew();
		AinsMakeStq(saveIns, r, REG_SP, AINS_DATA(iins));
		AINS_BBL(saveIns) = jbbl;
		AinsInsertAfter(saveIns, iins);
	    }
	    else {
		/* shrinking the stack */
		AINS_DATA(iins) += 8;
		/* add in a load of callee save register, since
		 *  we used it for computing values in the loop 
		 */
		/* stq r, (sp)  */
		saveIns = AinsNew();
		AinsMakeLdq(saveIns, r, REG_SP, AINS_DATA(iins));
		AINS_BBL(saveIns) = jbbl;
		AinsInsertBefore(saveIns, iins);
		
	    }
	}
    }
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "OptDupIndir"


/* @FUN
 * look for the following pattern in a loop: 
 * bne (after address computation) with both
 * branches being function calls, one a jsr, the other
 * a bsr.
 *
 * If we find that pattern, duplicate all blocks dominated by bne
 * block (except the two calls within the loop), and put one copy
 * under each call.  
 * @END 
 */
extern BOOL OptDupIndir( FILE *fp )
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "DupIndir");
    BOOL change = 0;
    INDEX ibbl;
    INDEX iins_bne;
    INDEX tbbl;
    INDEX fbbl;
    clock_t time = clock();
    NUMBER dcount = 0;
    INDEX ifun;
    REGLIST free_regs;
    INDEX jbbl;
    REG reg;
    INDEX iins;
    INDEX prebbl;
    INDEX iins2;
    INDEX newIns;
    NUMBER high, low;

    FOREACH_LIVE_BBL(ibbl)
    {
	if (ABBL_TYPE(ibbl) != BT_BRANCH) continue;

	if ( !ABBL_IS_HOT(ibbl) ) continue;

	iins_bne = ABBL_INS_LAST(ibbl);
	if (AINS_CODE(iins_bne) != I_bne) continue;

#if 0
	if (!ABBL_IS_IN_ANY_LOOP(ibbl)) 
	    printf("it is not in a loop\n");
#endif
        
	/* now we know we have a bne block, look at the two 
	   successors */
	tbbl = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
	fbbl = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
	
	if (verbose)
	    printf("looking at %d and %d\n", tbbl, fbbl);
	if ((ABBL_TYPE(tbbl) != BT_CALL) || (AbblNInss(tbbl) != 1)) 
	    continue;
	printf("it is a call instruction in the true branch\n");

	if ((ABBL_TYPE(fbbl) != BT_CALL) || (AbblNInss(fbbl) != 1)) 
	    continue;
	printf("it is a call instruction in the false branch\n");

	/* now we know both successors are calls, 
	   and have one instruction */
	if ((AINS_CODE(ABBL_INS_LAST(tbbl)) != I_bsr) && 
	    (AINS_CODE(ABBL_INS_LAST(fbbl)) != I_bsr)) 
	    continue;

	printf("one is a bsr\n");

	if ((AINS_CODE(ABBL_INS_LAST(tbbl)) != I_jsr) && 
	    (AINS_CODE(ABBL_INS_LAST(fbbl)) != I_jsr)) 
	    continue;
	printf("one is a jsr\n");

	/* look for a free callee saved register (one not used by
	 * current function 
	 */
	ifun = ABBL_FUN(ibbl);
	free_regs = SAVE_REGS;
	FOREACH_FUN_BBL(ifun, jbbl) {
	    FOREACH_BBL_INS(jbbl, iins) {
		/* check each assignment against a list of 
		   registers */
	        free_regs = free_regs & ~(REG2BIT(AINS_REG_A(iins)));
	        free_regs = free_regs & ~(REG2BIT(AINS_REG_B(iins)));
	        free_regs = free_regs & ~(REG2BIT(AINS_REG_C(iins)));
	    }
	}
	if (free_regs == REGS_NONE) {
	    /* if there are no calee save registers free, exit loop 
	     */
	    printf("no free regs, so bail\n");
	    continue;
	}
	printf("free_regs is %lx\n", free_regs);

	/* select any of the free callee saved registers, and steal it 
	 */
	reg = RegsGetFirst(free_regs);

	printf("selected register %d\n", reg);

	/* now find all instructions that subtract or add to the sp,
           and add 8 to the distance */

	AfunStealCalleeRegister(ifun, reg);
	
	printf("stole a register\n");
	dcount++;

	/* find loop pre-header */

	prebbl = ABBL_IS_LOOP_HEADER(ibbl) ? ibbl : ABBL_LOOP_HEADER(ibbl);
	prebbl = PreHeaderOf(prebbl);
	printf("header of loop is %d\n", prebbl);

	/* use the high and low from the instructions you will kill */
	/* now kill the two address computation instructions
	   in the loop, and use the current reg for the computation */
	iins = ABBL_INS_LAST(ibbl);
	
	iins = AINS_PREV(iins);
	AINS_REG_C(iins) = reg;
	
	iins = AINS_PREV(iins);
	iins2 = AINS_PREV(iins);
	AinsKill(iins);
	AinsKill(iins2);

	high = AINS_DATA(iins2);
	low = AINS_DATA(iins);
	
	/* insert a computation of the address before the 
	   loop starts */
	newIns = AinsNew();
	AinsMakeLdah(newIns,REG_GP, reg , high);
	AINS_BBL(newIns) = prebbl;
	AinsPrepend(newIns);

	newIns = AinsNew();
	AinsMakeLda(newIns, reg, reg, low);
	AINS_BBL(newIns) = prebbl;
	AinsPrepend(newIns);
	


    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (dcount>0);
    InfoStatus(fp,"\t[OptDupIndir%c] Inlined Funs: %d {%d ms}\n",
           CHANGE(change),dcount,time);

    return change;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "OptInlineIndir"

extern BOOL OptInlineIndir(FILE *fp )
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Indir");
    BOOL   change = 0;
    ADDRESS count;
    INDEX  i, j, index;
    INDEX iins, ibbl;
    INDEX targetBbl;
    INDEX targetfun;
    clock_t time = clock();
    NUMBER fcount = 0;
    INDEX targetReg, retReg;
    INDEX lastIns;
    INDEX jumpBbl;
    INDEX inlinedBbl;
    INDEX newIns;
    ADDRESS Y;
    NUMBER high, low;

#if 1
    return 0;
#endif

    if (( !CONFIG_USE_PROFILE ) || ( aIndir == 0 ) )
    {
        printf("\t[OptInlineIndir] No Inlining because of lack of profiling information\n");
        return 0;
    }
    

    /* only want to create a short new sequence, so regular inlining
     * will take care of the actual inlining 
     */
    
    for (i=0; i < aIndirFuns; i++) {
	/* skip this for indir calls with 0 as highest count */
	if (aIndir[i].values[0].count <= 0) continue;
	iins = aIndir[i].iins;
	ibbl = AINS_BBL(iins);

	/* if we've already resolved the jump, don't bother with this */
	if (AINS_CODE(ABBL_INS_LAST(ibbl)) != I_jsr) continue;

	printf("grabbed bbl %d, instruction %d\n", ibbl, iins);
	printf("function is %d\n", ABBL_FUN(ibbl));

	/* now find highest probability address */
	count = aIndir[i].values[0].count;
	index = 0;
	for (j = 1; j < CONFIG_NINDIR; j++) {
	    if (count < aIndir[i].values[j].count) {
		count = aIndir[i].values[j].count;
		index = j;
	    }
	    if (aIndir[i].values[j].count > 0) {
		printf("jumped to %lx %ld times\n", 
		       aIndir[i].values[j].addr, aIndir[i].values[j].count);
	    }
	}
	
	printf("most frequently executed indir jump is executed %ld/[%ld] times\n", aIndir[i].values[index].count, aIndir[i].totalcount);
	printf("most frequent percentage is %f\n",(double)aIndir[i].values[index].count/(double)(aIndir[i].totalcount)); 

	/* if the block is killed, or not hot, don't inline (duh!) */
	/* Also, most frequently executed branch must be > .66 to break 
	   even */
	if (ABBL_IS_KILLED(ibbl)) continue;
	if (!ABBL_IS_HOT(ibbl)) continue;
	if (((double)aIndir[i].values[index].count/(double)(aIndir[i].totalcount)) < .66) 
	    continue;
	
	targetBbl = AINS_BBL(Addr2Ins(aIndir[i].values[index].addr));
	targetfun = ABBL_FUN(targetBbl);
	
	if (AFUN_FLAGS(targetfun) & F_MARKED) {
	    if (verbose)
		printf("marked thingy\n");
	}

	if (AfunNBblsHot(targetfun) == 0) {
	    if (verbose)
		printf("no hot blocks\n");
	    continue;
	}
	if (AFUN_HAS_STK_INVALID(targetfun)) {
	    if (verbose)
		printf("invalid stack\n");
	}
	if ((AfunIsSetjmp(targetfun)) || (AfunIsLongjmp(targetfun))) {
	    if (verbose)
		printf("it is setjump/longjump\n");
	    continue;
	}
	if (AfunHasEsc2(targetfun)) {
	    if (verbose)
		printf("has escapes on the critical path\n");
	    continue;
	}

	fcount ++;

	/* now insert test for highest probability address */
	iins = ABBL_INS_LAST(ibbl);

	if (verbose) {
	    printf("iins is %d\n", iins);
	    PrintIns(stdout, iins); 
	    
	    printf("return register is %d\n", AINS_REG_C(iins));
	    printf("global pointer is %lx\n", SysGetGlobalPointer());
	    printf("starting search at %lx\n", SysGetGlobalPointer() - 32760);
	    printf("looking for address %lx\n", aIndir[i].values[index].addr);
	}

	lastIns = ABBL_INS_LAST(ibbl);
	retReg = AINS_REG_C(iins);
	targetReg = AINS_REG_B(iins);

	Y = aIndir[i].values[index].addr - SysGetGlobalPointer();
	high = Y >> 16;
	low = Y & 0xffff;
	if( low & 0x8000 )
	{
	    low |= 0xffff0000;
	    high += 1;
	}
#if 0
	WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
	CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
	
	newIns = AinsNew();
	AinsMakeLdah(newIns,REG_GP, retReg ,high);
	AINS_BBL(newIns) = ibbl;
	AinsInsertBefore(newIns, lastIns);

	newIns = AinsNew();
	AinsMakeLda(newIns, retReg, retReg , low);
	AINS_BBL(newIns) = ibbl;
	AinsInsertBefore(newIns, lastIns);

	newIns = AinsNew();
	AinsMakeSubq(newIns, retReg, targetReg, retReg);
	AINS_BBL(newIns) = ibbl;
	AinsInsertBefore(newIns, lastIns);
	ABBL_TYPE(ibbl) = BT_BRANCH;

	/* now put the branch in so we can bail if it's the wrong
           value */
	newIns = AinsNew();
	AinsMakeBne(newIns, retReg, 0);
	AINS_BBL(newIns) = ibbl;
	AinsInsertBefore(newIns, lastIns);


	/* split off the jsr from this bbl, and make it it's own bbl */

	jumpBbl = AbblNew();
	ABBL_FUN(jumpBbl) = ABBL_FUN(ibbl);
	AbblAppendFancy(jumpBbl);
	ABBL_OUT(jumpBbl) = ABBL_OUT(ibbl);

	AinsUnlink(lastIns);
	AINS_PREV(lastIns) = AINS_NEXT(lastIns) = INVALID;
	AINS_BBL(lastIns) = jumpBbl;
	AinsAppend(lastIns);

	ABBL_TYPE(jumpBbl) = BT_CALL;
	AedgLinkFancy( AedgNewFancy(jumpBbl, AEDG_TAIL(ABBL_EDG_LINK(ibbl)), 
				    ET_FUN_LINK));
	AedgLinkFancy( AedgNewFancy(jumpBbl, AEDG_TAIL(ABBL_EDG_CALL(ibbl)), 
				    ET_FUN_CALL));
	ABBL_COUNT(jumpBbl) = ABBL_COUNT(ibbl) - aIndir[i].values[index].count;

	/* make a bsr call, so the regular inlining stuff will take care 
	 * of this 
	 */
	inlinedBbl = AbblNew();
	ABBL_FUN(inlinedBbl) = ABBL_FUN(ibbl);
	ABBL_TYPE(inlinedBbl) = BT_CALL;
	ABBL_OUT(inlinedBbl) = ABBL_OUT(ibbl);
	ABBL_COUNT(inlinedBbl) = aIndir[i].values[index].count;
	AedgLinkFancy( AedgNewFancy(inlinedBbl,AEDG_TAIL(ABBL_EDG_LINK(ibbl)), 
				    ET_FUN_LINK));
	AedgLinkFancy( AedgNewFancy(inlinedBbl, AINS_BBL(Addr2Ins(aIndir[i].values[index].addr)), 
				    ET_FUN_CALL));
	AEDG_COUNT(ABBL_EDG_CALL(inlinedBbl)) =  aIndir[i].values[index].count;
	AEDG_FLAGS(ABBL_EDG_CALL(inlinedBbl)) |= FE_INLINABLE;

	if (verbose) {
	    printf("edg is %d, flags are %x\n", ABBL_EDG_CALL(inlinedBbl),
		   AEDG_FLAGS(ABBL_EDG_CALL(inlinedBbl)));
	    
	    printf("inlined to bbl %d\n", AINS_BBL(Addr2Ins(aIndir[i].values[index].addr)));
	    printf("function of bbl is %d [%s]\n", 
		   ABBL_FUN(AINS_BBL(Addr2Ins(aIndir[i].values[index].addr))),
		   AFUN_NAME(ABBL_FUN(AINS_BBL(Addr2Ins(aIndir[i].values[index].addr))))
		);
	    PrintFunSuper(stdout, ABBL_FUN(AINS_BBL(Addr2Ins(aIndir[i].values[index].addr))));
	}

	/* add this bbl to the function */
	AbblAppendFancy(inlinedBbl);

	iins = AinsNew();
	AinsMakeBsr(iins, retReg);
	AINS_BBL(iins) = inlinedBbl;
	AinsAppend(iins);


	/* kill off the edges that used to leave this bbl */
	AedgKillFancy(ABBL_EDG_CALL(ibbl));
	AedgKillFancy(ABBL_EDG_LINK(ibbl));

	/* set edges leaving the bbl */
	ABBL_TYPE(ibbl) = BT_BRANCH;
	AedgLinkFancy( AedgNewFancy(ibbl, jumpBbl, ET_TRUE) );
	AedgLinkFancy( AedgNewFancy(ibbl, inlinedBbl, ET_FALSE) );


	if (verbose) {
	    printf("ibbl is %d\n", ibbl);
	    PrintBbl(stdout, ibbl); 
	    PrintBbl(stdout, jumpBbl); 
	    PrintBbl(stdout, inlinedBbl); 
	printf("ibbl is %d, new jump bbl is %d, inlined jump bbl is %d\n", 
	       ibbl, jumpBbl, inlinedBbl);
	printf("count of inlined bbl is %ld\n", ABBL_COUNT(inlinedBbl));
	}
	printf("inlined (indirect) into function %s\n", AFUN_NAME(ABBL_FUN(ibbl)));
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (fcount>0);

    InfoStatus(fp,"\t[OptInlineIndir%c] Inlined Funs: %d {%d ms}\n",
           CHANGE(change),fcount,time);

    return change;
}


