/*=======================================================
  Author:   R. Muth
  Version:  $Id: printall.c,v 1.22 1998/11/23 04:13:02 muth Exp muth $
  Comment:  This file contains the high level printing code
=======================================================*/

#include "global.h"
#include "functions.h"
#include "flags.h"

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllStk"

extern void PrintAllStk(FILE *fp)
{
    INDEX ifun;

    FOREACH_LIVE_FUN(ifun)
    {
        if( !HasOption(AFUN_NAME(ifun),PRINT_OPT_FUN) ) continue;
        
        PrintFun(fp,ifun);
        PrintFunStk(fp,ifun);
    }
    
}


/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintAllLoop"

extern void PrintAllLoop(FILE *fp)
{
   INDEX ifun;

   ComputeEdgeCounts(fp);

   fprintf(fp, "\n%%StartSloop\n");

   FOREACH_LIVE_FUN(ifun)
   {
       if( HasOption(AFUN_NAME(ifun),PRINT_OPT_FUN) )
	  PrintFunLoops(fp, ifun, 0);
   }

   fprintf(fp, "\n%%EndSloop\n");

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllIns"

extern void PrintAllIns(FILE *fp)
{
    INDEX iins;

    FOREACH_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_BBL_BEG )
        {
            INDEX iedg;
            INDEX ibbl = AINS_BBL(iins);
            fprintf(fp,"%d [%c%c%c] %s ",
                    ibbl,
                    ABBL_FLAGS(ibbl) & F_KILLED ? ' ' : 'K',
                    ABBL_FLAGS(ibbl) & F_REFQUAD? ' ' : 'Q',
                    ABBL_FLAGS(ibbl) & F_GPREL32? ' ' : 'G',
                    AFUN_NAME(ABBL_FUN(ibbl)));

            FOREACH_BBL_PRED(ibbl,iedg)
            {
                fprintf(fp," %d",AEDG_HEAD(iedg));
            }
            fprintf(fp,":\n");
        }
        
        PrintIns(fp,iins);
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllFun"

extern void PrintAllFun(FILE *fp)
{
    INDEX ifun;

    FOREACH_FUN(ifun)
    {
        PrintFun(fp,ifun);
        PrintSep3(fp);
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllFunSuper"

extern void PrintAllFunSuper(FILE *fp)
{
    INDEX ifun;

    FOREACH_LIVE_FUN(ifun)
    {
        if( strcmp(AFUN_NAME(ifun),PRINT_OPT_FUN) ) continue;
        PrintFunSuper(fp,ifun);
    }

}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllBbl"

extern void PrintAllBbl(FILE *fp)
{
    INDEX ibbl;

    FOREACH_BBL(ibbl)
    {
        PrintBbl(fp,ibbl);
        PrintSep3(fp);
    }

}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllHotBbl"

extern void PrintAllHotBbl(FILE *fp)
{
    INDEX ibbl;

    FOREACH_LIVE_BBL(ibbl)
    {
        if( !ABBL_IS_HOT(ibbl) ) continue;
        PrintBblSuper(fp,ibbl);
        PrintSep3(fp);
    }

}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllEdg"

extern void PrintAllEdg(FILE *fp)
{
    INDEX iedg;

    FOREACH_EDG(iedg)
    {
        PrintEdg(fp,iedg);
    }

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintConfig"

extern void PrintConfig(FILE *fp)
{
    fprintf(fp,
            "Alto compiled " __DATE__ " " __TIME__ "\n"
            "Print What:  %s\n"
            "Print When:  %s\n" 
            "Print Fun:   %s\n"
            "Disabl. Opt: %s\n"
            "Count:       %6ld\n"
            "In File:     %s\n"
            "Out File:    %s\n"
            "Sections:    %6d\n"
            "Functions:   %6d\n"
            "Bblocks:     %6d\n"
            "Edges:       %6d\n"
            "Instr.:      %6d\n"
            "Stacks:      %6d\n"
            "Fun Align:   %6d\n"
            "Loop Align:  %6d\n"
            "Bbl Align:   %6d\n",
            PRINT_OPT_WHAT,
            PRINT_OPT_WHEN,
            PRINT_OPT_FUN,
            DISABLED_OPT,
            CONFIG_COUNT,
            CONFIG_INFILE,
            CONFIG_OUTFILE,
            CONFIG_NSEC,
            CONFIG_NFUN,
            CONFIG_NBBL,
            CONFIG_NEDG,
            CONFIG_NINS,
            CONFIG_NSTK,
            CONFIG_FUN_ALIGN,
            CONFIG_LOOP_ALIGN,
            CONFIG_BBL_ALIGN);
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintUsedResources"

extern void PrintUsedResources(FILE *fp)
{
    fprintf(fp,
            "Functions: %6d/%6d\n"
            "Bblocks:   %6d/%6d\n"
            "Edges:     %6d/%6d\n"
            "Instr.:    %6d/%6d\n",
            CONFIG_UFUN,
            CONFIG_NFUN,
            CONFIG_UBBL,
            CONFIG_NBBL,
            CONFIG_UEDG,
            CONFIG_NEDG,
            CONFIG_UINS,
            CONFIG_NINS);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAll"

extern void PrintAll(FILE *fp)
{
    /*  */
    
    ComDomsAndLoops(fp);

    if( HasOption(PRINT_OPT_WHAT,"Config") )
        PrintConfig(fp);

    if( HasOption(PRINT_OPT_WHAT,"Used") )
        PrintUsedResources(fp);
    
    /*  */
    
    if( HasOption(PRINT_OPT_WHAT,"Head") )
        PrintAllHead(fp);

    if( HasOption(PRINT_OPT_WHAT,"Data") )
        PrintAllData(fp);

    if( HasOption(PRINT_OPT_WHAT,"Code") )
        PrintAllCode(fp);

    if( HasOption(PRINT_OPT_WHAT,"Rel") )
        PrintAllReloc(fp);
#if 0
    if( HasOption(PRINT_OPT_WHAT,"Fdr") )
        PrintAllFdr(fp);
    
    if( HasOption(PRINT_OPT_WHAT,"Pdr") )
        PrintAllPdr(fp);
#endif
    if( HasOption(PRINT_OPT_WHAT,"Sym") )
        PrintAllSym(fp);

    if( HasOption(PRINT_OPT_WHAT,"Loop") )
    {
	OptRemoveUnreachableCode(fp);
#if 1
        ComDomsAndLoops(fp);
#endif        
        PrintAllLoop(fp);
    }
    
    if( HasOption(PRINT_OPT_WHAT,"Stk") )
        PrintAllStk(fp);

    if( HasOption(PRINT_OPT_WHAT,"Fun") )
        PrintAllFun(fp);

    if( HasOption(PRINT_OPT_WHAT,"Bbl") )
        PrintAllBbl(fp);

    if( HasOption(PRINT_OPT_WHAT,"Hotbbl") )
        PrintAllHotBbl(fp);

    if( HasOption(PRINT_OPT_WHAT,"Ins") )
        PrintAllIns(fp);

    if( HasOption(PRINT_OPT_WHAT,"Edg") )
        PrintAllEdg(fp);

    /* checking data structures */

    if( HasOption(PRINT_OPT_WHAT,"Super") )
        PrintAllFunSuper(fp);

    if( HasOption(PRINT_OPT_WHAT,"Statbin") )
        PrintStatBinary(fp);

    if( HasOption(PRINT_OPT_WHAT,"Statstk") )
        PrintStatStack(fp);

        

    if( HasOption(PRINT_OPT_WHAT,"Statprof") )
        PrintProfile(fp);
    
    /* checking data structures */
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

}

/*=======================================================
=======================================================*/
