/*=======================================================
  Author:   
  Version:  $Id$
  Comment:  Instrumentation code
=======================================================*/



#include "global.h"
#include "functions.h"
#include "flags.h"
#include <stdio.h>
#include "opcode.h"
#include "regs.h"

#ifdef SYS_COFF
/* don't include these on linux-alpha */
#include <cmplrs/prof_header.h>
#include <cmplrs/pdf.h>
#include "system.coff.h"
#endif

#ifdef SYS_ELF
#include "system.elf.h"
#endif

static long insMapAddress = 0;

#define CALL_PROF_SIZE 42*8
#define START_PROF 1
#define AINS_USE_DEF_POTENTIAL(i) AINS(i).Potential


/*=======================================================
  =======================================================*/
#undef __PROC__
#define __PROC__ "ComInsertProfilingInformation"

/* Insert a count at each basic block.  Assumes we have a fully
 * constructed flow graph (including function entries, etc.)
 */



/* counter space goes like this
 * counter1
 * counter2
 * ....
 * countern
 * spill space
 * .data section
 */
extern void ComInsertProfilingInformation(FILE *fp)
{
    ADDRESS counterStart;
    NUMBER bbls=0;
    NUMBER edges=0;
    ADDRESS counterSpace = 0;
    NUMBER funs = 0;
    NUMBER jumps = 0;
    NUMBER iins = 0;
    NUMBER values = 0;
    NUMBER counter;
    INDEX ibbl;
    INDEX sweepFunBbl;

    if (!HasOption(CONFIG_PROFILE_WHAT, "Value"))
	ProfStrRedIndirectCalls(); 
    counterSpace = (CONFIG_UBBL + 2*CONFIG_UEDG) * 8 + CONFIG_UINS*4;
    
    
    /* find out how many counters we need for calls and jumps */
    counter = 0;
    FOREACH_BBL(ibbl) {
	if ((ABBL_TYPE(ibbl) == BT_JUMP) && 
	    (HasOption(CONFIG_PROFILE_WHAT, "Jump")))
	    counter++;
	if ((ABBL_TYPE(ibbl) == BT_CALL) &&
	    (HasOption(CONFIG_PROFILE_WHAT, "Call")))
	    counter++;
    };


    if (HasOption(CONFIG_PROFILE_WHAT, "Value")) {
	counter += CountVPCode(fp);
    }
    counterSpace += counter*CALL_PROF_SIZE;


    /* round up */
    counterSpace = (counterSpace + 0xffff) & (~0xffff);
    printf("counter space is %lx \n", counterSpace);
    printf("data section starts at %lx\n", ASEC_START(BIN_SECTION_MAP[R_SN_DATA]));
    IncreaseData(counterSpace);


    counterStart = ASEC_START(BIN_SECTION_MAP[R_SN_DATA]) + 5*8;

    /* each bbl has a counter, even bogus ones */
    if (HasOption(CONFIG_PROFILE_WHAT, "Bbl"))
	bbls =  ComInsertBBProfilingInformation(counterStart, fp);
    counterStart += bbls*8;
    printf("counterstart is %lx\n", counterStart);

    /* all edges will be profiled */
    if (HasOption(CONFIG_PROFILE_WHAT, "Edg"))
	edges = ComInsertEdgeProfilingInformation(counterStart, fp);
    counterStart += 8*(edges*2);
    printf("counterstart is %lx\n", counterStart);

    sweepFunBbl = ComInsertValueProfiling(fp);

    if (HasOption(CONFIG_PROFILE_WHAT, "Call"))
	funs = ComInsertCallProfilingInformation(counterStart, sweepFunBbl, fp); 
    counterStart += funs*(CALL_PROF_SIZE);
    printf("funs is %d\n", funs);

    printf("counterstart is %lx\n", counterStart);

    if (HasOption(CONFIG_PROFILE_WHAT, "Jump"))
	jumps = ComInsertJmpProfilingInformation(counterStart, sweepFunBbl, fp); 

    funs += jumps;
    printf("funs is %d\n", funs);
    counterStart += jumps*(CALL_PROF_SIZE);

    printf("counterstart is %lx\n", counterStart);

#ifdef SYS_ELF    
    CFATAL( ((bbls + 2*edges)*8 > CONFIG_COUNTERSPACE), 
	    (" NOT ENOUGH SPACE FOR COUNTERS!!! need %lx\n", (bbls + 2*edges) *8));
    printf("need %lx space for counters, allocated %lx\n", (bbls+2*edges)*8, CONFIG_COUNTERSPACE);
    
#endif
    
    if (HasOption(CONFIG_PROFILE_WHAT, "Value")) {
	values = ComInsertValueProfilingInformation(counterStart, sweepFunBbl, fp);
    }

    funs += values;
    printf("funs is %d\n", funs);
    counterStart += values*(CALL_PROF_SIZE);
    printf("counterstart is %lx\n", counterStart);

    iins = ComGetInsMapCount();
    printf("size of instruction map is %d\n", iins);
    insMapAddress = counterStart;

    /* output must print all counters in order */
    if ((HasOption(CONFIG_PROFILE_WHAT, "Bbl")) || 
	(HasOption(CONFIG_PROFILE_WHAT, "Edg"))) {
	ComInsertProfilingOutput(bbls, edges, funs, iins, 
				 ASEC_START(BIN_SECTION_MAP[R_SN_DATA]), 
				 "AltoCounts", fp);
	
    }
    
    else {
	ComInsertProfilingOutput(bbls, edges, funs, iins, 
				 ASEC_START(BIN_SECTION_MAP[R_SN_DATA]), 
				 "ValueCounts", fp);
    }
	

}


extern int ComGetInsMapCount()
{
    INDEX iins;
    NUMBER count = 0;

    FOREACH_INS(iins) {
	if (AINS_IS_RELOC(iins)) {
	    count++;
	}
    }
    
    return count;
}


extern void ComInsertInsMap(FILE *fp)
{
    INDEX iins;
    NUMBER count = 0;
    ADDRESS map = insMapAddress;


    FOREACH_INS(iins) {
	if (AINS_IS_RELOC(iins)) {
	    WriteData64(map, Ins2Addr(iins));
	    WriteData64(map + 8, TranslateAddress(Ins2Addr(iins)));

#if 0	    
	    printf("writing at address %lx\n", map);
	    printf("wrote values %lx and %lx\n", ReadData64(map), ReadData64(map+8));
#endif
	    map += 16;
	    count++;
	}
    }
}



/* we will us 41 words per counter, 1 quad, then 40 in pairs.  Each
 * pair will contain an address, and a count for that address. The
 * first quad will contain the address of the jump instruction.
 */
extern int ComInsertCallProfilingInformation(ADDRESS counterStart, INDEX sweepFunBbl, FILE *fp)
{
    INDEX ibbl;
    NUMBER count = 0;
    INDEX lastIns;
    INDEX iins;
    ADDRESS Y;
    NUMBER high, low;
    ADDRESS counterEnd;
    INDEX ifun;
    INDEX startBbl;
    INDEX resetBbl;
    INDEX callsiteBbl;
    REG prof1, prof2, prof3;
    REG target;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Callprof");
    verbose = 1;

    /* first count the bbls with calls */
    FOREACH_LIVE_BBL(ibbl) {
	if (ABBL_TYPE(ibbl) != BT_CALL) continue;
	if (AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA ) continue;
	count++;
    }
    
    if (verbose)
	fprintf(fp, "total of %d calls profiled\n", count);
    

    counterEnd = counterStart + count * CALL_PROF_SIZE;


    count = 0;
    
    FOREACH_LIVE_BBL_R(ibbl) {
	if (ABBL_TYPE(ibbl) != BT_CALL) continue;
	if (AINS_TYPE(ABBL_INS_LAST(ibbl)) == IT_BRA ) continue;
	
	/* ok, here's the indirect function call */
	lastIns = ABBL_INS_LAST(ibbl);
	ifun = ABBL_FUN(ibbl);

	/* these are needed for the function call, do not use free
	   registers.  We can store them in a free register instead
	   of the stack if we need to, but that my be overkill.
	   */

	/* at some point we can change these to be free registers */
	prof1 = REG_A0;
	prof2 = REG_A1;
	prof3 = REG_RA;
	target = AINS_REG_B(lastIns);

	if (verbose) {
	    fprintf(fp, "ibbls is %d\n", ibbl);
	    fprintf(fp, "profiling at instruction %d (%lx)\n", lastIns,
		    Ins2Addr(lastIns));
	    
	    fprintf(fp,"using register %d to jump through\n", AINS_REG_B(lastIns));
	    fprintf(fp,"count is %d\n", count);
	}

	/* ok, now add a some basic blocks to handle table lookups */
	ibbl = AbblNew();
	ABBL_FUN(ibbl) = ifun;
	AbblAppendFancy(ibbl);
	startBbl = ibbl;


	/* dec sp by 24 (enough to store 3 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, -24);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* stq $a0, 8(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof1, REG_SP, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* stq $a1, 0(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof2, REG_SP, 8);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* stq $a1, 0(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof3, REG_SP, 16);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);
	
	/* do the actual counter manipulations here */
	Y = counterEnd - (count + 1)*(CALL_PROF_SIZE);

	if (verbose) {
	    fprintf(fp, "starting counter at %lx, count is %d\n", Y, count);
	    fprintf(fp, "address profiled is %lx\n", Ins2Addr(lastIns));
	}

	/* store the instruction profiled */
	WriteData64(Y, Ins2Addr(lastIns));

	/* store the counter for value profiling */
	WriteData64(Y+8, START_PROF);

	/* store the -1 target value as a catchall */
	WriteData64(Y + (CALL_PROF_SIZE - 16), -1);

	Y += 8;  /* move past the address and counter */
	Y -= SysGetGlobalPointer();


	high = Y >> 16;
	low = Y & 0xffff;
	if( low & 0x8000 )
	{
	    low |= 0xffff0000;
	    high += 1;
	}
#if 0
	WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
	CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
	
	/* set the register to be profiled to be a1 */
	iins = AinsNew();
	AinsMakeLda(iins, target, prof2, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);


	/* set the t1 to counter address */
	iins = AinsNew();
	AinsMakeLdah(iins,REG_GP, prof1 ,high);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	iins = AinsNew();
	AinsMakeLda(iins,prof1, prof1 ,low);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);



	/* call the function */
	iins = AinsNew();
	AinsMakeBsr(iins, prof3);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* ok, call bbl is in place */

	/* restore when we return */

	ibbl = AbblNew();
	ABBL_FUN(ibbl) = ifun;
	AbblAppendFancy(ibbl);
	resetBbl = ibbl;

	/* ldq $ra, 8(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof3, 16);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* ldq $a0, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof2, 8);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);


	/* ldq $a0, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof1, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* reset stack */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP , 24);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);


	/* this is the real call site */
	callsiteBbl = AbblNew();
	ABBL_FUN(callsiteBbl) = ifun;
	AbblAppendFancy(ibbl);

	ibbl = AINS_BBL(lastIns);

	/* remove from old bbl */
	AinsUnlink(lastIns);
        AINS_PREV(lastIns) = AINS_NEXT(lastIns) = INVALID;

	/* add to new bbl */
	AINS_BBL(lastIns) = callsiteBbl;
	AinsAppend(lastIns);


	/* now set up links to new blocks */
	ABBL_TYPE(startBbl) = BT_CALL;
	AedgLinkFancy(AedgNewFancy(startBbl, resetBbl, ET_FUN_LINK));
	AedgLinkFancy(AedgNewFancy(startBbl, sweepFunBbl, ET_FUN_CALL));


	ABBL_TYPE(resetBbl) = BT_NORM;
	AedgLinkFancy(AedgNewFancy(resetBbl, callsiteBbl, ET_NORM));

	ABBL_TYPE(callsiteBbl) = BT_CALL;
        AedgLinkFancy(AedgNewFancy(callsiteBbl,AEDG_TAIL(ABBL_EDG_LINK(ibbl)), 
                                    ET_FUN_LINK));
        AedgLinkFancy(AedgNewFancy(callsiteBbl,AEDG_TAIL(ABBL_EDG_CALL(ibbl)), 
                                    ET_FUN_CALL));


        /* kill off the edges that used to leave this bbl */
        AedgKillFancy(ABBL_EDG_CALL(ibbl));
        AedgKillFancy(ABBL_EDG_LINK(ibbl));

	/* set this bbl */
	ABBL_TYPE(ibbl) = BT_NORM;
	AedgLinkFancy(AedgNewFancy(ibbl, startBbl, ET_NORM));


	/* print all the blls now to see if they match up correctly */
	if (verbose) {
	    printf("FFFFFFFFFOOOOOOO\n");
	    PrintBblSuper(fp, ibbl);
	    PrintBblSuper(fp, startBbl);
	    PrintBblSuper(fp, resetBbl);
	    PrintBblSuper(fp, callsiteBbl);
	}
	count++;
    }

    InfoStatus(fp,"\t[InsertCallProfiling] total: %d\n", count);
    return count;
}


/* we will us 41 words per counter, 1 quad, then 40 in pairs.  Each
 * pair will contain an address, and a count for that address. The
 * first quad will contain the address of the jump instruction.
 */
extern int ComInsertJmpProfilingInformation(ADDRESS counterStart, INDEX sweepFunBbl, FILE *fp)
{
    INDEX ibbl;
    NUMBER count = 0;
    INDEX lastIns;
    INDEX iins;
    ADDRESS Y;
    NUMBER high, low;
    INDEX iins_jmp, iins_add, iins_ldl, iins_ind;
    REG indexReg, target;
    INDEX prev;
    REG reg;
    ADDRESS counterEnd;
    INDEX ifun;
    INDEX startBbl;
    INDEX resetBbl;
    INDEX callsiteBbl;
    REG prof1, prof2, prof3;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Jumpprof");

    if (verbose)
	fprintf(fp, "Profiling Jumps Now\n");
    /* these should all be in switch blocks, so find the index register */

    /* first count the bbls with calls */
    FOREACH_LIVE_BBL(ibbl) {
	if (ABBL_TYPE(ibbl) != BT_JUMP) continue;
	if (AINS_TYPE(ABBL_INS_LAST(ibbl)) != IT_JMP ) continue;
	
	count++;
    }
    
    if (verbose)
	fprintf(fp, "total of %d jumps profiled\n", count);
    

    counterEnd = counterStart + count * CALL_PROF_SIZE;

    if (verbose)
	fprintf(fp, "counters start at %lx and end at %lx\n", counterStart, counterEnd);

    fflush(stdout);
#if 0
    return 0;
#endif

    count = 0;
    
    FOREACH_LIVE_BBL_R(ibbl) {
	if (ABBL_TYPE(ibbl) != BT_JUMP) continue;
	if (AINS_TYPE(ABBL_INS_LAST(ibbl)) != IT_JMP ) continue;
	prof1 = REG_A0;
	prof2 = REG_A1;
	prof3 = REG_RA;

	/* these are needed for the function call, do not use free
	   registers.  We can store them in a free register instead
	   of the stack if we need to, but that my be overkill.
	   */
	ifun = ABBL_FUN(ibbl);

	
	/* ok, here's the indirect function call */
	lastIns = ABBL_INS_LAST(ibbl);

	if (verbose) {
	    fprintf(fp, "profiling jump at instruction %d (%lx)\n", lastIns,
		    Ins2Addr(lastIns));
	    
	    fprintf(fp, "using register %d to jump through\n", AINS_REG_B(lastIns));
	}

	prev = ABBL_PREV(ibbl);
	/* find JMP instructions ==================================*/
	
	iins_jmp = FindTableJmp(ibbl,prev);
	if( !VALID(iins_jmp) )
	{
	    ERROR(("invalid iins_jmp\n"));
	}


	/* find ADD instructions =======================*/
	
	iins_add = FindTableAdd(ibbl,prev,iins_jmp);
	if( !VALID(iins_add) )
	{
	    ERROR(("invalid iins_add\n"));
	}
        
	/* find LDL instructions ==================================*/
	
	iins_ldl = FindTableLdl(ibbl,prev,iins_add);
	if( !VALID(iins_ldl) )
	{
	    ERROR(("invalid iins_ldl\n"));
	}
	
	
	/* find IND instructions ==========================*/
	
	reg = AINS_REG_B(iins_ldl);
	for( iins_ind=AINS_PREV(iins_ldl); VALID(iins_ind); iins_ind=AINS_PREV(iins_ind))
	{
	    if( AINS_REG_C(iins_ind) == reg ) break;
	}
    
	if( !VALID(iins_ind) )
	    FOREACH_BBL_INS_R(prev,iins_ind)
	    {
		if( AINS_REG_C(iins_ind) == reg ) break;
	    }
	
	if( !VALID(iins_ind) ||
	    AINS_CODE(iins_ind)!=I_s4addq)
	{
	    INDEX ifun = ABBL_FUN(ibbl);
	    WARNING(("could not find ind for bt_jump %d in %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
	}
	
	/* at this point I've found the index instr, load instr, and
	   add instr */

	/* I essentially need to check the index to see if it's going
	 * to load the correct jump table entry */
	if (VALID(iins_jmp) && VALID(iins_ldl) && 
	    VALID(iins_add) && VALID(iins_ind)) {
	    indexReg = AINS_REG_A(iins_ind);
	}
	else {
	    indexReg = AINS_REG_B(lastIns);
	}

#if 0
	if (indexReg == prof1) {
	    prof1 = REG_T3;
	}
	else if (indexReg == prof2) {
	    prof2 = REG_T3;
	}
#endif

	if (verbose) {
	    printf("profiling register register is %d\n", indexReg);
	    printf("using registers %d and %d to profile\n", prof1, prof2);
	    fprintf(fp, "ibbls is %d\n", ibbl);
	    fprintf(fp, "profiling at instruction %d (%lx)\n", lastIns,
		    Ins2Addr(lastIns));
	    
	    fprintf(fp,"using register %d to jump through\n", AINS_REG_B(lastIns));
	    fprintf(fp,"count is %d\n", count);
	}

	target = indexReg;

	/* ok, now add a some basic blocks to handle table lookups */
	ibbl = AbblNew();
	ABBL_FUN(ibbl) = ifun;
	AbblAppendFancy(ibbl);
	startBbl = ibbl;


	/* dec sp by 16 (enough to store 2 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, -24);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* stq $a0, 0(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof1, REG_SP, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* stq $a1, 8(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof2, REG_SP, 8);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);
	

	/* stq $ra, 16(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, prof3, REG_SP, 16);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);
	
	/* do the actual counter manipulations here */
	Y = counterEnd - (count + 1)*(CALL_PROF_SIZE);

	if (verbose) {
	    fprintf(fp, "starting counter at %lx, count is %d\n", Y, count);
	    fprintf(fp, "address profiled is %lx\n", Ins2Addr(lastIns));
	}

	/* store the instruction profiled */
	WriteData64(Y, Ins2Addr(lastIns));

	/* store the counter for value profiling */
	WriteData64(Y+8, START_PROF);

	/* store the -1 target value as a catchall */
	WriteData64(Y + (CALL_PROF_SIZE - 16), -1);

	Y += 8;  /* move past the address and counter */
	Y -= SysGetGlobalPointer();


	high = Y >> 16;
	low = Y & 0xffff;
	if( low & 0x8000 )
	{
	    low |= 0xffff0000;
	    high += 1;
	}
#if 0
	WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
	CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
	
	/* set the register to be profiled to be a1 */
	iins = AinsNew();
	AinsMakeLda(iins, target, prof2, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);



	/* set the t1 to counter address */
	iins = AinsNew();
	AinsMakeLdah(iins,REG_GP, prof1 ,high);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	iins = AinsNew();
	AinsMakeLda(iins,prof1, prof1 ,low);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);



	/* call the function */
	iins = AinsNew();
	AinsMakeBsr(iins, prof3);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* ok, call bbl is in place */

	/* restore when we return */

	ibbl = AbblNew();
	ABBL_FUN(ibbl) = ifun;
	AbblAppendFancy(ibbl);
	resetBbl = ibbl;

	/* ldq $ra, 8(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof3, 16);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* ldq $a0, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof2, 8);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);


	/* ldq $a0, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, prof1, 0);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* reset stack */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP , 24);
	AINS_BBL(iins) = ibbl;
	AinsAppend(iins);

	/* this is the real jump site */
 	callsiteBbl = AbblNew();
	ABBL_FUN(callsiteBbl) = ifun;
	AbblAppendFancy(callsiteBbl);
	ABBL_TYPE(callsiteBbl) = BT_JUMP;

	ibbl = AINS_BBL(lastIns);

	AinsUnlink(lastIns);
        AINS_PREV(lastIns) = AINS_NEXT(lastIns) = INVALID;

	AINS_BBL(lastIns) = callsiteBbl;
	AinsAppend(lastIns);



	/* now set up links to new blocks */
	ABBL_TYPE(startBbl) = BT_CALL;
	AedgLinkFancy(AedgNewFancy(startBbl, resetBbl, ET_FUN_LINK));
	AedgLinkFancy(AedgNewFancy(startBbl, sweepFunBbl, ET_FUN_CALL));


	ABBL_TYPE(resetBbl) = BT_NORM;
	AedgLinkFancy(AedgNewFancy(resetBbl, callsiteBbl, ET_NORM));

        /* kill off the edges that used to leave this bbl */
        AedgKillFancy(ABBL_EDG_JUMP(ibbl));

	AedgLinkFancy( AedgNewFancy(callsiteBbl, AFUN_BBL_FIRST(AFUN_PSEUDO_HELL) , ET_JUMP ) );


	/* set this bbl */
	ABBL_TYPE(ibbl) = BT_NORM;
	AedgLinkFancy(AedgNewFancy(ibbl, startBbl, ET_NORM));


	/* print all the blls now to see if they match up correctly */
	if (verbose) {
	    printf("FFFFFFFFFOOOOOOO\n");
	    PrintBblSuper(fp, ibbl);
	    PrintBblSuper(fp, startBbl);
	    PrintBblSuper(fp, resetBbl);
	    PrintBblSuper(fp, callsiteBbl);
	}

	count++;
    }


    InfoStatus(fp,"\t[InsertJumpProfiling] total: %d\n", count);
    return count;
}


/* we will take up 2 words per counter.  The first word is the index of the 
 * tail, and index of the head of the edge.  The second word is the actual 
 * counter 
 * we also need to figure out what to do will call site information
 */

#define TRUE_ONLY 1

extern int ComInsertEdgeProfilingInformation(ADDRESS counterStart, FILE *fp)
{
    INDEX iins, ibbl;
    NUMBER profcount = 0;
    INDEX profbbl;
    ADDRESS Y;
    NUMBER high, low;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Edgecount");
    INDEX  iedg;
    ADDRESS counterEnd;

    if( verbose )
        fprintf(fp, "\t[*]starting edge counters at %lx\n", (long) counterStart);


    FOREACH_EDG_R(iedg) {
#if TRUE_ONLY
	if (AEDG_TYPE(iedg) != ET_TRUE)
	    continue;
#else
	if ((AEDG_TYPE(iedg) != ET_TRUE) && 
	    (AEDG_TYPE(iedg) != ET_FALSE)) 
	    continue;
#endif
	profcount ++;
    }
    counterEnd = counterStart + 2*8*profcount;

    if( verbose )
    {
        fprintf(fp,"\t[*] total counters is %d\n", profcount);
        fprintf(fp,"\t[*] ending edge counters at %lx\n", counterEnd);
    }
    
    profcount = 0;

    FOREACH_EDG_R(iedg)
    {
#if TRUE_ONLY
	if (AEDG_TYPE(iedg) != ET_TRUE)
	    continue;
#else
	if ((AEDG_TYPE(iedg) != ET_TRUE) && 
	    (AEDG_TYPE(iedg) != ET_FALSE)) 
	    continue;
#endif

	if (verbose)
        {
	    fprintf(fp,"\t[*] profiling edge %d\n", iedg);
	    if (AEDG_TYPE(iedg) == ET_TRUE)
            {
		fprintf(fp,"\t[*] edge is a true edge\n");
	    }
	}
	/* do exact same steps for only true branch */
	ibbl = AEDG_FROM(iedg);

	profbbl = AbblNew();
	ABBL_TYPE(profbbl) = BT_NORM;
	ABBL_FUN(profbbl) = ABBL_FUN(AEDG_HEAD(iedg));
	AbblAppendFancy(profbbl);
	
	/* dec sp by 16 (enough to store 2 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, -16);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	if (verbose)
	    fprintf(fp, "\t[*] adding instructions to bbl %d\n", profbbl);
	/* ok, spill gp and scratch register to stack */
	
	/* stq $t2, 0(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T2, REG_SP, 0);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* stq $t1, 8(sp) */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T1, REG_SP, 8);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* now add the counters to this bbl */
	
	Y = counterEnd - 2*8*(profcount+1);
	if (verbose)
        {
	    fprintf(fp, "\t[*] counter is at %lx\n", Y);
	}


	WriteData64(Y, iedg);
	Y+= 8;

	
	Y -= SysGetGlobalPointer();
	
	high = Y >> 16;
	low = Y & 0xffff;
	if( low & 0x8000 )
	{
	    low |= 0xffff0000;
	    high += 1;
	}
#if 0
	WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
	CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
	
	iins = AinsNew();
	AinsMakeLdah(iins,REG_GP, REG_T2 ,high);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* ldq $t1, low($t2) */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_T2, REG_T1, low);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	
	/* addq $t1, 0x1, $t1 */
	iins = AinsNew();
	AinsMakeLda(iins, REG_T1, REG_T1, 1);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* stq $t1, low($t2) */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T1, REG_T2, low);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* ok, load gp and spill register from stack */
	/* ldq $gp, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, REG_T2, 0);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* ldq $t1, 8(sp) */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, REG_T1, 8);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	/* inc sp by 16 (enough to store 2 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, 16);
	AINS_BBL(iins) = profbbl;
	AinsAppend(iins);
	
	
	if (verbose)
        {
	    fprintf(fp, "\t[*] adding edge to block %d, ins %d to %d\n", profbbl, ABBL_INS_FIRST(profbbl), ABBL_INS_LAST(profbbl));
	    fprintf(fp, "\t[*] old bbl is %d \n", ibbl);
	}
	/* now move the old edge to point to this guy, and 
	 * add a fall-through edge to next block */
	AedgKillFancy( iedg );
	AedgLinkFancy( AedgNewFancy(AEDG_HEAD(iedg), profbbl,
				    AEDG_TYPE(iedg)) );
	AedgLinkFancy( AedgNewFancy(profbbl, AEDG_TAIL(iedg), ET_NORM) );

	if (verbose)
	    PrintBblSuper(fp, profbbl);

	profcount ++;

    }
    
    InfoStatus(fp,"\t[InsertEdgeProfiling] total: %d\n", profcount);

    return(profcount);
}


extern int ComInsertBBProfilingInformation(ADDRESS counterStart, FILE *fp)
{
    ADDRESS Y;
    NUMBER high, low;
    INDEX i, iins;
    INDEX firstIns;
    INDEX entryBbl;
    BOOL verbose = 0;

#if 1
    fprintf(fp, "currently %d bbls\n", CONFIG_UBBL);
    fprintf(fp, "passing program entry, entry is %lx\n", SysGetProgramEntry());
#endif
    
    iins = Addr2Ins(SysGetProgramEntry());
    entryBbl = AINS_BBL(iins);


    FOREACH_LIVE_BBL(i)
    {
	if (!VALID(ABBL_INS_FIRST(i))) continue;
	if (i != entryBbl) {
	    firstIns = ABBL_INS_FIRST(i);
	}
	else {
	    firstIns = ABBL_INS_FIRST(i);
	    firstIns = AINS_NEXT(firstIns);
	    firstIns = AINS_NEXT(firstIns);
	    firstIns = AINS_NEXT(firstIns);
	}

	if (verbose) {
	    fprintf(fp, "address of counter is %lx\n", counterStart + i*8);
	    fprintf(fp, "first instruction is %d\n", firstIns);
	}

	/* store the old global pointer and the compute register 
	 *  on the stack (3 instructions) 
	 */
	/* dec sp by 16 (enough to store 2 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, -16);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* stq $t1, 8(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T1, REG_SP, 8);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* stq $t2, 0(sp)  */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T2, REG_SP, 0);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);



	/* set the global pointer to counter address */
	Y = counterStart + i*8;
	Y -= SysGetGlobalPointer();

	high = Y >> 16;
	low = Y & 0xffff;
	if( low & 0x8000 )
	{
	    low |= 0xffff0000;
	    high += 1;
	}
#if 0
	WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
	CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
	
	iins = AinsNew();
	AinsMakeLdah(iins,REG_GP, REG_T2 ,high);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

        /* ldq $t1, low($t2) */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_T2, REG_T1, low);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* addq $t1, 0x1, $t1 */
	iins = AinsNew();
	AinsMakeLda(iins, REG_T1, REG_T1, 1);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* stq $t1, low($t2) */
	iins = AinsNew();
	AinsMakeStq(iins, REG_T1, REG_T2, low);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* ok, load gp and spill register from stack */
	/* ldq $t2, 0(sp)  */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, REG_T2, 0);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* ldq $t1, 8(sp) */
	iins = AinsNew();
	AinsMakeLdq(iins, REG_SP, REG_T1, 8);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

	/* inc sp by 16 (enough to store 2 registers) */
	iins = AinsNew();
	AinsMakeLda(iins, REG_SP, REG_SP, 16);
	AINS_BBL(iins) = i;
	AinsInsertBefore(iins, firstIns);

    }
   
    return CONFIG_UBBL;
}


/* 
 * As arguments, we take number of counters, and start of counters.
 */

extern void ComInsertProfilingOutput(NUMBER bblcounters, NUMBER edgcounters, NUMBER funcounters, NUMBER insMap, ADDRESS datasection, STRING outname, FILE *fp)
{
    INDEX ibbl, iins;
    int i = 0;
    NUMBER counters = bblcounters + 2*edgcounters + (CALL_PROF_SIZE/8)*funcounters + 2*insMap;
    ADDRESS counterEnd = 8*counters + 5*8;
    char *location;
    ADDRESS Y;
    NUMBER high, low;
    INDEX FirstIns;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"ProfOut");

    location = (char *)ASEC_DATA8(BIN_SECTION_MAP[R_SN_DATA]);
    
    ((long *)location)[0] =  ALTO_MAGIC;
    ((long *)location)[1] =  bblcounters;
    ((long *)location)[2] =  edgcounters;
    ((long *)location)[3] =  funcounters;
    ((long *)location)[4] =  insMap;

    fprintf(fp, "magic %lx, bbls %lx, edgs %lx, funs %lx, insmap %lx\n", ((long *)location)[0],((long *)location)[1],((long *)location)[2],((long *)location)[3], ((long *)location)[4]);


    fprintf(fp, "writing count output file name to data section at address %lx\n", (long)counterEnd);
    location += counterEnd;  /* magic number + #bbl + #edg + #fun + #insmap */
    sprintf(location, "%s.%s", CONFIG_INFILE, outname);
    fprintf(fp, "number of counters is %d\n", counters);
    fprintf(fp, "address of counter start is %lx\n", datasection);
    

    iins = 0; ibbl = 0;
    /* patch _exit function to call the above basic block */
    /* first we need to find the _exit symbol....*/
    FOREACH_FUN(i) {
	if (strcmp("_exit", AFUN_NAME(i)) == 0) {
	    /* is the name of the symbol _exit */
	    printf("found _exit function, index  is %d\n", i);
	    FOREACH_FUN_BBL(i, ibbl) {
		if (ABBL_TYPE(ibbl) == BT_PAL) {
		    printf("bbl is %d\n", ibbl);
		    goto outer;  /* don't bother with the rest of the loop */
		}
	    }
	}
    }
  outer:

    printf("ibbl is %d\n", ibbl);
    FirstIns = ABBL_INS_FIRST(ibbl);
    printf("instruction is %d\n", FirstIns);
    printf("address of last instruction is %lx\n", Ins2Addr(ABBL_INS_FIRST(ibbl)));

    /* set the global pointer to file name address */
    Y = counterEnd + datasection;
    printf("counter end is %lx\n", Y);
    Y -= SysGetGlobalPointer();
    printf("offset from gp is %lx\n", Y);
    
    high = Y >> 16;
    low = Y & 0xffff;
    if( low & 0x8000 )
    {
	low |= 0xffff0000;
	high += 1;
    }
#if 0
    WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
    CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
    
    /* put start address in A0 */
    iins = AinsNew();
    AinsMakeLdah(iins,REG_GP, REG_A0,high);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);
    
    iins = AinsNew();
    AinsMakeLda(iins,REG_A0, REG_A0,low);    
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);
	

    /* lda        a1,1537(zero) (flags)                */
    iins = AinsNew();
    Disassemble(0x223f0601, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /* lda        a2, 384(zero) (flags) */
    iins = AinsNew();
    AinsMakeLda(iins, REG_ZERO, REG_A2, 384);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);
     

    /* lda        v0,45(zero)   (open call)            */
    iins = AinsNew();
    Disassemble(0x201f002d, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /* callsys                     */
    iins = AinsNew();
    Disassemble(0x00000083, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /*  mov        v0,t0         (save file pointer)   */
    iins = AinsNew();
    Disassemble(0x44000401, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /* now move address of beginning of counters into gp */
    Y = datasection - SysGetGlobalPointer();
    printf("global pointer is %lx\n", SysGetGlobalPointer());
    printf("datasection counter start is %lx\n", datasection);
    printf("counter start is %lx\n", Y);
    high = Y >> 16;
    low = Y & 0xffff;
    if( low & 0x8000 )
    {
	low |= 0xffff0000;
	high += 1;
    }
#if 0
    WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
    CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
    
    /* put the address of the counter start in a1 (address of filename is in a0) */
    iins = AinsNew();
    AinsMakeLdah(iins,REG_GP, REG_A1 ,high);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);
    
    iins = AinsNew();
    AinsMakeLda(iins,REG_A1, REG_A1,low);    
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /*  mov        t0,a0         (file point is 1st arg)  */
    iins = AinsNew();
    Disassemble(0x44210410, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);


    /* put the number of bytes to write in a2 */
    Y = (counters + 5)*8; /* 8 bytes per counter plus */
    printf("space of counters is %lx\n", Y);
    high = Y >> 16;
    low = Y & 0xffff;
    if( low & 0x8000 )
    {
	low |= 0xffff0000;
	high += 1;
    }
#if 1
    WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif

    iins = AinsNew();
    AinsMakeLdah(iins,REG_ZERO, REG_A2, high);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    iins = AinsNew();
    AinsMakeLda(iins,REG_A2, REG_A2, low);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /*  lda        v0,4(zero)       */
    iins = AinsNew();
    Disassemble(0x201f0004, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /* callsys                     */
    iins = AinsNew();
    Disassemble(0x00000083, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /*  mov        t0,a0            */
    iins = AinsNew();
    Disassemble(0x44210410, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    /*  lda        v0,6(zero)       */
    iins = AinsNew();
    Disassemble(0x201f0006, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);


    /* callsys                     */
    iins = AinsNew();
    Disassemble(0x00000083, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, FirstIns);

    if (verbose) 
	PrintBblSuper(stdout, ibbl);

}

